// editor '93
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static void UnInitEditor()
// Undo what InitEditor did
{
	UnInitUserAbort();
	UnInitClipboard();
	UnInitBuffers();
	UnInitSyntaxMaps();
	UnInitKeyBindingTable();
}

static bool InitEditor()
// Initialize all high-level editor concepts
// If there is a problem, set the error and return false.
{
	if(InitKeyBindingTable())
	{
		if(InitSyntaxMaps())
		{
			if(InitBuffers())
			{
				if(InitClipboard())
				{
					if(InitUserAbort())
					{
						return(true);
					}
					else
					{
						SetError("Failed to initialize user abort");
					}
					UnInitClipboard();
				}
				else
				{
					SetError("Failed to initialize clipboard");
				}
				UnInitBuffers();
			}
			else
			{
				SetError("Failed to initialize buffers");
			}
			UnInitSyntaxMaps();
		}
		else
		{
			SetError("Failed to initialize syntax maps");
		}
		UnInitKeyBindingTable();
	}
	else
	{
		SetError("Failed to initialize key bindings");
	}
	return(false);
}

bool InitMain(int argc, const char* argv[])
// The e93 entry point
{
	bool
		fail;
	UINT32
		numPointers;
	size_t
		stringLength;
	const char
		*leafName;

	fail=false;
	/* 2018-01-20, gab: When called from applicationDidFinishLaunching (AppDeletate.m), argv is
	 * provided in a malloc'd block of memory in an autorelease pool, which disappears when
	 * that method terminates. Therefore, we have to COPY the programName out of that block of
	 * memory into our own, persistent, block of malloc'd memory.
	 */
	// find leaf name to use as programName
	leafName = strrchr(argv[0], '/');
	if (leafName == NULL)
		leafName = (char *) argv[0];		// if no separator, just use the name as-is
	else
		leafName += 1;	// skip past the separator
	stringLength = strlen(leafName) + 1;		// +1 for the terminating \0
	if ((programName = (char *) malloc(stringLength)) == NULL)
	{
		fprintf(stderr, "Failed to allocate memory for programName");
		return(false);
	}
	strncpy(programName, leafName, stringLength);		// point to the program name forever more
	setlocale(LC_ALL,"");								// adjust to local customs
	if(EarlyInit())										// give process a chance to do whatever it would like (NOTE: if it returns false, exit without complaint (so it can fork if it wants to))
	{
		if(InitErrors())
		{
			ClearErrorTrace();
			if(InitEnvironment())						// handle low level GUI and environment init
			{
				ClearErrorTrace();
				if(InitEditor())						// handle high level initialization
				{
					if (ShellLoopInit(argc, argv))
					{
						return(true);
					}
				}
				else
				{
					fprintf(stderr,"Failed to initialize editor: %s\n",GetErrorTrace());
					fail=true;
				}
				UnInitEnvironment();
			}
			else
			{
				fprintf(stderr,"Failed to initialize environment: %s\n",GetErrorTrace());
				fail=true;
			}
			UnInitErrors();
		}
		else
		{
			fprintf(stderr,"Could not initialize error handlers\n");
			fail=true;
		}
		EarlyUnInit();
	}
	if((numPointers=MGetNumAllocatedPointers()))
	{
		fprintf(stderr,"Had %d pointer(s) allocated on exit\n",numPointers);
		fail=true;
	}
	return(!fail);
}

void UninitMain()
// The e93 entry point
{
	UnInitEditor();
	UnInitEnvironment();
	UnInitErrors();
	EarlyUnInit();
}
