// Low level event handling
// Copyright (C) 2009 Fred Allen.
//
// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include "includes.h"
#import <Cocoa/Cocoa.h>


#define kCaretTimer			0.562
#define kUpdateTimer		0.050

@interface EventHandlerObject : NSObject
@end

@implementation EventHandlerObject
{
}

- (id)init
{
	if (self = [super init])
	{
	}
	return self;
}

- (void)caretTimer:(NSTimer*)timer
{
	WINDOWLISTELEMENT* theWindowElement;

	if ((theWindowElement = LocateActiveEditorWindow()))
	{
		EDITOR_WINDOW* theWindow;

		theWindow = theWindowElement->theEditorWindow;
		BlinkViewCursor((EDITOR_VIEW*)theWindow->windowInfo);		// blink active windows cursor
	}
}

- (void)timerTimer:(NSTimer*)timer
{
	static UINT32 onceASecTimer = 0;					// timer goes off every 50ms, start counter at zero

	onceASecTimer++;
	if (onceASecTimer >= (1000 / 50))
	{
		onceASecTimer = 0;
		HandleShellCommand("timer", 0, NULL);			// service editor shell timer once a ~second
	}
	UpdateBufferTasks();
}

@end

static EventHandlerObject* eventHandlerObject;
static NSTimer* caretTimer;
static NSTimer* timerTimer;
static id hotKeyEventHandler;


UINT32 OSXToEditorModifiers(NSEventModifierFlags modifiers)
// Convert OS X modifiers to editors
{
	UINT32 evMod;

	evMod = 0;
	if (modifiers & NSEventModifierFlagShift)				// shift pressed?
	{
		evMod |= EEM_SHIFT;									// yes
	}
	if (modifiers & NSEventModifierFlagControl)				// control?
	{
		evMod |= EEM_CTL;									// yes
	}
	if (modifiers & NSEventModifierFlagCommand)				// command key
	{
		evMod |= EEM_MOD0;									// yes
	}
	if (modifiers & NSEventModifierFlagOption)				// option key
	{
		evMod |= EEM_MOD1;									// yes
	}
	return (evMod);
}

void ResetCursorBlinkTime(EDITOR_VIEW* theView)
// Kill the timer that blinks the Cursor, and set it up fresh, if the view is active.
{
	if (caretTimer != nil)
	{
		[caretTimer invalidate];
		caretTimer = nil;
	}
	if (theView->active)
	{
		caretTimer = [NSTimer scheduledTimerWithTimeInterval:kCaretTimer target:eventHandlerObject selector:@selector(caretTimer:) userInfo:nil repeats:YES];
	}
}


static void UninstallOSXHandlers()
// Uninstall all OS X handlers
{
	[NSEvent removeMonitor:hotKeyEventHandler];
	if (timerTimer != nil)
	{
		[timerTimer invalidate];
		timerTimer = nil;
	}
	if (caretTimer != nil)
	{
		[caretTimer invalidate];
		caretTimer = nil;
	}
}

static bool InstallOSXHandlers()
// Install all the OS X handlers
{
	eventHandlerObject = [[EventHandlerObject alloc] init];
	caretTimer = [NSTimer scheduledTimerWithTimeInterval:kCaretTimer target:eventHandlerObject selector:@selector(caretTimer:) userInfo:nil repeats:YES];
	timerTimer = [NSTimer scheduledTimerWithTimeInterval:kUpdateTimer target:eventHandlerObject selector:@selector(timerTimer:) userInfo:nil repeats:YES];

	hotKeyEventHandler = [NSEvent addLocalMonitorForEventsMatchingMask:NSKeyDownMask handler:
	^(NSEvent* incomingEvent)
	{
		NSEvent* result = incomingEvent;
		// control break key pressed
		if ([incomingEvent type] == NSKeyDown && [incomingEvent keyCode] == 113 && ([incomingEvent modifierFlags] & NSEventModifierFlagControl))
		{
			g_hotKeyPressed = true;
			result = nil;
		}
		return result;
	}];
	return (true);
}

INT32 FindOSXVirtualKey(INT32 theKey)
{
	int i;

	for (i = 0; i < 256; i++)
	{
		if (g_menuVirtKeyArray[i] == theKey)
		{
			return (i);
		}
	}
	return (-1);
}

bool EditorKeyNameToKeyCode(const char* theName, UINT32* theCode)
// given a string which is the name of an unmodified key on the keyboard,
// convert it into a keycode, if the name is not the name of any key on our keyboard, then return FALSE
{
	UINT32 i;
	bool found;

	found = false;
	for (i = 0; !found && (i < 256); i++)
	{
		if (strcasecmp(theName, g_keyNameArray[i]) == 0)
		{
			found = true;
			(*theCode) = i;
		}
	}
	return (found);
}

char* EditorKeyCodeToKeyName(UINT32 theCode)
// given a keycode, get its name
// if the code is invalid, return NULL
// NOTE: the returned string is static, and should not be modified or deleted
{
	return ((char*)g_keyNameArray[theCode & 0xFF]);
}

INT32 EditorKeyCodeToVirtKeyCode(UINT32 theCode)
// given a keycode, get its virtual key code or 0
{
	return (g_menuVirtKeyArray[theCode & 0xFF]);
}

bool EditorGetKeyPress(UINT32* keyCode, UINT32* editorModifiers, bool wait, bool clearBuffered)
// Used by editor when it wants to get a key press
// If a code is found, it is returned in keyCode/editorModifiers
// If none is ready, return false
// If wait is true, wait here until one is ready (true is always returned)
// If clearBuffered is true, then clear any keypresses that may have been in the
// buffer before proceeding.
{
	bool results;
	BOOL keepOn = YES;

	results = false;
	while (keepOn)
	{
		NSEvent* theEvent = [NSApp
							 nextEventMatchingMask:NSKeyDownMask
							 untilDate:[[NSDate alloc] initWithTimeIntervalSinceNow:0.02]
							 inMode:NSDefaultRunLoopMode
							 dequeue:YES];

		if ([theEvent type] == NSEventTypeKeyDown)
		{
			*editorModifiers = OSXToEditorModifiers((UINT32)[theEvent modifierFlags]);										// convert OS X modifiers to editors
			*keyCode = theEvent.keyCode;
			keepOn = NO;
			results = true;
		}
	}
	return results;
}

void EditorQuit()
// This is called from the editor when it wants to quit
// it needs to inform the event loop that it should stop looping
{
	[NSApp performSelector:@selector(terminate:) withObject:nil afterDelay:0.1];
}

bool EditorEventLoop(int argc, const char* argv[])
// Get the next event that the editor needs to know about and return it
// if other events occur, this routine should stay here handling them until
// one specifically related to the editor arrives, it should then pass that
// event up to the editor for processing
// if no events arrive, it is the responsibility of this routine to
// give time back to the OS if necessary.
{
	if (argc > 1 && strstr(argv[1], "-NS"))						// when e93 is launch from the Finder, argv[1] is the process number in the form of -psn-0-xxxxxxx
	{
		HandleShellCommand("initialargs", argc - 2, &(argv[2]));	// pass off initial arguments to shell for processing, but skip argv[1]
	}
	else
	{
		HandleShellCommand("initialargs", argc - 1, &(argv[1]));	// pass off initial arguments to shell for processing, skipping past argv[0] always
	}
	Tcl_SetServiceMode(TCL_SERVICE_ALL);
	return true;
}

void ClearAbort()
// call this to clear any pending abort status
{
	g_hotKeyPressed = false;
}

bool CheckAbort()
// This will return true if the user is requesting an abortion of the
// current job
// This should be made to execute as efficiently as possible, since it will
// be called often during certain loops
// To make it more efficient, this version checks the time, and if enough time
// has not passed, returns instantly
{
	return (g_hotKeyPressed);
}

void UnInitEvents()
// Uninitialize event stuff
{
	UninstallOSXHandlers();
}

bool InitEvents()
// Initialize event stuff
// if there is a problem, SetError, return false
{
	g_hotKeyPressed = false;
	if (InstallOSXHandlers())
	{
		return (true);
	}
	return (false);
}
