// Font handling
// Copyright (C) 2009 Fred Allen.
//
// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include "includes.h"
#import <Cocoa/Cocoa.h>

UniChar static const MAC_ROMAN_TO_UNICODE[256] =
{
	0, 1, 2, 3, 4, 5, 6, 7,
	8, 9, 10, 11, 12, 13, 14, 15,
	16, 17, 18, 19, 20, 21, 22, 23,
	24, 25, 26, 27, 28, 29, 30, 31,
	32, 33, 34, 35, 36, 37, 38, 39,
	40, 41, 42, 43, 44, 45, 46, 47,
	48, 49, 50, 51, 52, 53, 54, 55,
	56, 57, 58, 59, 60, 61, 62, 63,
	64, 65, 66, 67, 68, 69, 70, 71,
	72, 73, 74, 75, 76, 77, 78, 79,
	80, 81, 82, 83, 84, 85, 86, 87,
	88, 89, 90, 91, 92, 93, 94, 95,
	96, 97, 98, 99, 100, 101, 102, 103,
	104, 105, 106, 107, 108, 109, 110, 111,
	112, 113, 114, 115, 116, 117, 118, 119,
	120, 121, 122, 123, 124, 125, 126, 127,
	196, 197, 199, 201, 209, 214, 220, 225,
	224, 226, 228, 227, 229, 231, 233, 232,
	234, 235, 237, 236, 238, 239, 241, 243,
	242, 244, 246, 245, 250, 249, 251, 252,
	8224, 176, 162, 163, 167, 8226, 182, 223,
	174, 169, 8482, 180, 168, 8800, 198, 216,
	8734, 177, 8804, 8805, 165, 181, 8706, 8721,
	8719, 960, 8747, 170, 186, 937, 230, 248,
	191, 161, 172, 8730, 402, 8776, 8710, 171, 187,
	8230, 160, 192, 195, 213, 338, 339, 8211, 8212,
	8220, 8221, 8216, 8217, 247, 9674, 255, 376, 8260,
	8364, 8249, 8250, 64257, 64258, 8225, 183, 8218, 8222,
	8240, 194, 202, 193, 203, 200, 205, 206, 207,
	204, 211, 212, 63743, 210, 218, 219, 217,
	305, 710, 732, 175, 728, 729, 730, 184, 733, 731, 711
};


#define XLFD_FOUNDRY		0
#define XLFD_FAMILY			1
#define XLFD_WEIGHT			2
#define XLFD_SLANT			3
#define XLFD_SETWIDTH		4
#define XLFD_ADD_STYLE		5
#define XLFD_PIXEL_SIZE		6
#define XLFD_POINT_SIZE		7
#define XLFD_RESOLUTION_X	8
#define XLFD_RESOLUTION_Y	9
#define XLFD_SPACING		10
#define XLFD_AVERAGE_WIDTH	11
#define XLFD_CHARSET		12
#define XLFD_NUMFIELDS		13	// Number of fields in XLFD.

struct StateMap
{
	int keyValue;				// Integer representation of a value.
	const char* strKey;			// String representation of a value.
};


//
// The following structures are used when parsing XLFD's into a set of
//

static struct StateMap xlfdWeightMap[] =
{
	{0,									"normal"},
	{0,									"medium"},
	{0,									"book"},
	{NSFontCondensedTrait,				"extralight"},
	{NSFontCondensedTrait,				"light"},
	{NSFontBoldTrait,					"bold"},
	{NSFontBoldTrait,					"demi"},
	{NSFontBoldTrait,					"demibold"},
	{NSFontBoldTrait,					"extrabold"},
	{0,									NULL}			// Assume anything else is "normal".
};


static struct StateMap xlfdSlantMap[] =
{
	{0,					"r"},
	{NSFontItalicTrait,	"i"},
	{0,					"o"},
	{0, NULL}						// Assume anything else is "roman".
};


static EDITOR_FONT* fontListHead = NULL;

static int FindStateNum(char* stateStr, struct StateMap* theState)
// map a string to a number in the table.
// if stateStr was equal to the string keys of one of the elements in the
// theState table, returns the numeric key of that element.
{
	int i;
	bool done;

	i = 0;
	done = false;
	while (!done && theState[i].strKey)
	{
		if (strcmp(stateStr, theState[i].strKey) == 0)
		{
			done = true;
		}
		else
		{
			i++;
		}
	}
	return (theState[i].keyValue);
}

void LinkFontItem(EDITOR_FONT* theFont)
// passed a EDITOR_FONT, link it to the start of the global list
{
	if (fontListHead)														// point previous back to new
	{
		fontListHead->previous = theFont;
	}
	theFont->next = fontListHead;												// point to the old head of the list
	theFont->previous = NULL;													// no previous element
	fontListHead = theFont;													// point the head at the new element
}

void UnlinkFontItem(EDITOR_FONT* theFont)
// passed a EDITOR_FONT, unlink it from the global list
{
	if (theFont->next)
	{
		theFont->next->previous = theFont->previous;
	}
	if (theFont->previous)
	{
		theFont->previous->next = theFont->next;
	}
	else
	{
		fontListHead = theFont->next;											// this one was at the head, so update
	}
}


void GetFontInfo(const char* theFontName, EDITOR_FONT* theFont)
//	Passed in a editor font string, pull the info about the font out
{
	char* fields[XLFD_NUMFIELDS + 2];
	char fontStr[1024];
	int i;
	int numFields;
	int overhang;

	if (strlen(theFontName) > 0)
	{
		strcpy(fontStr, theFontName);
		strcpy(theFont->name, theFontName);
	}
	else
	{
		strcpy(fontStr, "-*-Arial-normal-r-normal--18--75-75-*-70-*-*");
		strcpy(theFont->name, "-*-Arial-normal-r-normal--18--75-75-*-70-*-*");
	}
	theFont->maxLeftOverhang = 0;	// initialize the overhang amounts
	theFont->maxRightOverhang = 0;
	for (i = 0, numFields = 0; fontStr[i] != '\0' && numFields <= XLFD_NUMFIELDS; i++)							// break the XLFD font string up into multiple strings
	{
		if (fontStr[i] == '-')
		{
			fontStr[i] = '\0';
			fields[numFields] = &(fontStr[i + 1]);														// point to each string for easy access
			numFields++;
		}
	}
	if ((theFont->size = atoi(fields[XLFD_PIXEL_SIZE])) == 0)											// use pixel size, but if it's zero
	{
		if ((theFont->size = atoi(fields[XLFD_POINT_SIZE])) != 0)										// use the point size if it's not zero
		{
			theFont->size = (theFont->size / 10);														// assume 72 DPI, should really be (theFont->size=MulDiv(theFont->size/10,deviceDPI,72);)

		}
		else
		{
			theFont->size = 16;																		// point and pixel sizes were zero, default to 16 for pixel size
		}
	}
	strncpy(theFont->family, fields[XLFD_FAMILY], sizeof(theFont->family));
	strncpy(theFont->weight, fields[XLFD_WEIGHT], sizeof(theFont->weight));
	strncpy(theFont->slant, fields[XLFD_SLANT], sizeof(theFont->slant));

	theFont->traits = FindStateNum(fields[XLFD_SLANT], xlfdSlantMap) | FindStateNum(fields[XLFD_WEIGHT], xlfdWeightMap);


	NSString* tmpStr = [NSString stringWithCString:fields[XLFD_FAMILY] encoding:NSASCIIStringEncoding];
	NSFontDescriptor* fontDesc = [NSFontDescriptor fontDescriptorWithFontAttributes:@ {NSFontFamilyAttribute:tmpStr}];
	NSFontDescriptor* fontTraitsDesc = [fontDesc fontDescriptorWithSymbolicTraits:theFont->traits];

	/* 2018-01-16, gab: If [NSFont fontWithDescriptor:size] can't find the named font on the system,
	 * it will return a nil reference. In that case, it's really not a good idea to continue to
	 * access the desired font information through that nil reference...
	 */
	if ((theFont->fontRef = [NSFont fontWithDescriptor:fontTraitsDesc size:theFont->size]) != nil) {
		strncpy(theFont->postScriptName, [theFont->fontRef.fontDescriptor.postscriptName cStringUsingEncoding:NSUTF8StringEncoding], sizeof(theFont->postScriptName) - 1);

		theFont->ascent = ceil(theFont->fontRef.ascender);													// make sure to round up to the nearest int
		theFont->descent = ceil(-theFont->fontRef.descender);
		theFont->leading = ceil(theFont->fontRef.leading);
		theFont->maxCharWidth = 0;

		for (i = 0; i < 256; i++)
		{
			UniChar c = MAC_ROMAN_TO_UNICODE[i];
			NSGlyph theGlyph = [theFont->fontRef glyphWithName:[NSString stringWithFormat:@"%C", c]];

			NSSize advance = [theFont->fontRef advancementForGlyph:theGlyph];
			NSRect gRect = [theFont->fontRef boundingRectForGlyph:theGlyph];

			// handle left overhang
			overhang = abs((INT32)ceil(gRect.origin.x));
			if (overhang > theFont->maxLeftOverhang)
			{
				theFont->maxLeftOverhang = overhang;
			}

			// handle right overhang
			overhang = (INT32)(ceil((gRect.size.width + gRect.origin.x) - advance.width));
			if (overhang > theFont->maxRightOverhang)
			{
				theFont->maxRightOverhang = overhang;
			}
			theFont->charWidths[i] = ceil(advance.width);
			if (theFont->charWidths[i] > theFont->maxCharWidth)
			{
				theFont->maxCharWidth = theFont->charWidths[i];
			}
		}	// for
	}	// if fontRef was located
}

void GetEditorFontName(EDITOR_FONT* font, char* nameBuffer, UINT32 bufferLength)
// Copy the name of the passed font into nameBuffer which is
// bufferLength bytes long (including the 0 termination)
{
	strncpy(nameBuffer, font->name, bufferLength);
}

void FreeEditorFont(EDITOR_FONT* theFont)
// Free a font allocated by LoadFont
{
	if (!(--theFont->usageCount))
	{
		if (theFont->previous)
		{
			theFont->previous->next = theFont->next;
		}
		else
		{
			fontListHead = theFont->next;
		}
		if (theFont->next)
		{
			theFont->next->previous = theFont->previous;
		}
		MDisposePtr(theFont);
	}
}

static EDITOR_FONT* LocateLoadedFont(const char* theName)
// see if an already loaded font can be located which has theName
// if so, return a pointer to it, otherwise return NULL
{
	EDITOR_FONT* testFont;

	testFont = fontListHead;
	while (testFont && (strcasecmp(testFont->name, theName) != 0))
	{
		testFont = testFont->next;
	}
	return (testFont);
}

EDITOR_FONT* LoadEditorFont(const char* theName)
// attempt to load the font given by name
// if there is a problem, return NULL
{
	EDITOR_FONT* theResult;

	if ((theResult = LocateLoadedFont(theName)))			// see if we can find this one already loaded
	{
		theResult->usageCount++;						// add another user to this font structure
	}
	else
	{
		if ((theResult = (EDITOR_FONT*)MNewPtrClr(sizeof(EDITOR_FONT))))	// allocate one
		{
			GetFontInfo(theName, theResult);
			theResult->usageCount = 1;
		}
	}
	return (theResult);
}

