// Copyright (C) 2009 Fred Allen.
// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include "includes.h"
#import <AppKit/AppKit.h>


static UINT32 modifierFlagTable[] =
{
	EEM_CAPSLOCK,
	EEM_SHIFT,
	EEM_CTL,
	EEM_MOD0,
	EEM_MOD1,
	EEM_MOD2,
	EEM_MOD3,
	EEM_MOD4,
	EEM_MOD5,
	EEM_MOD6,
	EEM_MOD7
};

static char localErrorFamily[] = "osxguimenus";

enum
{
	BADRELATIONSHIP,
	BADPATH,
	BADKEYNAME,
	BADMODIFIERS
};

static const char* errorMembers[] =
{
	"BadRelationship",
	"BadPath",
	"BadKeyName",
	"BadModifiers"
};

static const char* errorDescriptions[] =
{
	"Invalid menu relationship",
	"Invalid menu path",
	"Invalid key name",
	"Invalid modifier"
};


static EDITOR_MENU* firstMenu;
static UINT32 globalMenuID;

static EDITOR_MENU* LocateMenuWithItem(EDITOR_MENU* topLevel, UINT32 menuID)
// attempt to locate an active menu below topLevel, that has menuID defined
// if one can be located, return it, else return NULL
{
	EDITOR_MENU* currentMenu;
	EDITOR_MENU* locatedMenu;

	currentMenu = GetRelatedEditorMenu(topLevel, MR_FIRSTCHILD);
	locatedMenu = NULL;
	while (currentMenu && !locatedMenu)
	{
		if (!(locatedMenu = LocateMenuWithItem(currentMenu, menuID)))			// recurse to try to find child that has this key
		{
			if (currentMenu->menuID == menuID)									// see if this item has the key
			{
				locatedMenu = currentMenu;
			}
		}
		currentMenu = GetRelatedEditorMenu(currentMenu, MR_NEXTSIBLING);
	}
	return (locatedMenu);
}


@interface MenuObjCObject : NSObject
@end

@implementation MenuObjCObject
{
}

- (id)init
{
	if (self = [super init])
	{
	}
	return self;
}


- (void)handleMenuEvent:(NSMenuItem*)menu
{
	EDITOR_MENU* theMenu;

	switch (menu.tag)
	{
		case 2:
		{
			ExecuteTclCommand("AboutBox");				// handle default about menu
			break;
		}
		case 3:
		{
			HandleShellCommand("quit", 0, NULL);			// handle default quit menu
			break;
		}
		default:
		{
			if ((theMenu = LocateMenuWithItem(NULL, (UINT32)menu.tag)))
			{
				HandleMenuEvent(theMenu);									// call the shell to handle this menu
			}
		}
	}
}

@end

static MenuObjCObject* menuHandler;

EDITOR_MENU* GetRelatedEditorMenu(EDITOR_MENU* theMenu, int menuRelationship)
// return the menu with menuRelationship to theMenu
// return NULL if there is no menu with menuRelationship
{
	EDITOR_MENU* outputMenu;

	outputMenu = NULL;
	switch (menuRelationship)
	{
		case MR_NEXTSIBLING:
		{
			if (theMenu->parent)
			{
				if (theMenu->nextSibling != theMenu->parent->firstChild)
				{
					outputMenu = theMenu->nextSibling;
				}
			}
			else
			{
				if (theMenu->nextSibling != firstMenu)
				{
					outputMenu = theMenu->nextSibling;
				}
			}
			break;
		}
		case MR_PREVIOUSSIBLING:
		{
			if (theMenu->parent)
			{
				if (theMenu != theMenu->parent->firstChild)
				{
					outputMenu = theMenu->previousSibling;
				}
			}
			else
			{
				if (theMenu != firstMenu)
				{
					if (theMenu->previousSibling != firstMenu)
					{
						outputMenu = theMenu->previousSibling;
					}
				}
				else
				{
					outputMenu = firstMenu;
				}
			}
			break;
		}
		case MR_FIRSTCHILD:
		{
			if (theMenu)
			{
				outputMenu = theMenu->firstChild;
			}
			else
			{
				outputMenu = firstMenu;
			}
			break;
		}
		case MR_LASTCHILD:
		{
			if (theMenu)
			{
				if (theMenu->firstChild)
				{
					outputMenu = theMenu->firstChild->previousSibling;
				}
			}
			else
			{
				if (firstMenu)
				{
					outputMenu = firstMenu->previousSibling;
				}
			}
			break;
		}
		default:
		{
			break;
		}
	}
	return (outputMenu);
}

bool GetEditorMenu(int argc, const char* argv[], EDITOR_MENU** theMenu)
// traverse the elements of the menu path given by argc,argv
// and attempt to locate the menu that they point to
// if the path is invalid, return FALSE
// otherwise return TRUE with theMenu set to the located item
// NOTE: on entry, theMenu is used as the starting point for
// the search, if it is passed as NULL, it indicates that the search
// should start at the top
{
	EDITOR_MENU* currentMenu;
	EDITOR_MENU* testMenu;
	int elementIndex;
	bool matching;

	currentMenu = testMenu = *theMenu;
	matching = true;
	for (elementIndex = 0; (elementIndex < argc) && matching; elementIndex++)
	{
		if (currentMenu)											// locate root of current menu
		{
			testMenu = currentMenu->firstChild;
		}
		else
		{
			testMenu = firstMenu;
		}
		matching = false;
		while (testMenu && !matching)
		{
			if (strcmp(argv[elementIndex], testMenu->itemText) == 0)
			{
				matching = true;									// located a match, this is the menu we want
			}
			else
			{
				testMenu = GetRelatedEditorMenu(testMenu, MR_NEXTSIBLING);
			}
		}
		currentMenu = testMenu;
	}
	*theMenu = testMenu;
	return (matching);
}

void DeactivateEditorMenu(EDITOR_MENU* theMenu)
// make theMenu inactive
// if theMenu contains sub-Menus, they should be inactive because
// their parent is now inactive, but they should not have their active
// state changed. This way, when the parent becomes active again, the children
// will resume their old states
{
	theMenu->menuHandle.enabled = false;
}

void ActivateEditorMenu(EDITOR_MENU* theMenu)
// make theMenu active
{
	theMenu->menuHandle.enabled = true;
}

char* GetEditorMenuName(EDITOR_MENU* theMenu)
// return a pointer to the name of theMenu
{
	return (theMenu->itemText);
}

bool SetEditorMenuName(EDITOR_MENU* theMenu, const char* theName)
// set the name for theMenu
// if this fails, leave the old name, set the error, and return FALSE
{
	char* newName;

	if ((newName = (char*)MNewPtr((UINT32)(strlen(theName) + 1))))
	{
		strcpy(newName, theName);
		MDisposePtr(theMenu->itemText);
		theMenu->itemText = newName;
	}
	theMenu->menuHandle.title = [NSString stringWithCString:theName encoding:NSASCIIStringEncoding];
	return (true);
}

char* GetEditorMenuAttributes(EDITOR_MENU* theMenu)
// return a pointer to the attributes string for theMenu
{
	return (0);
}

bool SetEditorMenuAttributes(EDITOR_MENU* theMenu, const char* theAttributes)
// set the name for theMenu
// if this fails, leave the old name, set the error, and return FALSE
{
	return (true);
}

static void LinkEditorMenuElement(EDITOR_MENU* theMenu, EDITOR_MENU* baseMenu, int menuRelationship)
// link theMenu to baseMenu with menuRelationship
// if baseMenu is NULL, then menuRelationship must be MR_FIRSTCHILD, or MR_LASTCHILD as it is not possible to
// link menus as siblings of the root
// NOTE: passing baseMenu as NULL, with menuRelationship other than MR_FIRSTCHILD, or MR_LASTCHILD is
// _not_ detected, and will almost certainly cause bad things to happen
{
	theMenu->firstChild = NULL;
	switch (menuRelationship)
	{
		case MR_NEXTSIBLING:
		{
			theMenu->nextSibling = baseMenu->nextSibling;
			theMenu->nextSibling->previousSibling = theMenu;
			theMenu->previousSibling = baseMenu;
			baseMenu->nextSibling = theMenu;
			theMenu->parent = baseMenu->parent;
			break;
		}
		case MR_PREVIOUSSIBLING:
		{
			theMenu->previousSibling = baseMenu->previousSibling;
			theMenu->previousSibling->nextSibling = theMenu;
			theMenu->nextSibling = baseMenu;
			baseMenu->previousSibling = theMenu;
			theMenu->parent = baseMenu->parent;
			if (theMenu->parent->firstChild == baseMenu)
			{
				theMenu->parent->firstChild = theMenu;						// make us the first child if linked before old first child
			}
			break;
		}
		case MR_FIRSTCHILD:
		{
			theMenu->parent = baseMenu;										// set parent
			if (baseMenu)													// see if linking as a child of a current menu
			{
				if ((theMenu->nextSibling = baseMenu->firstChild))
				{
					theMenu->previousSibling = theMenu->nextSibling->previousSibling;
					theMenu->nextSibling->previousSibling = theMenu;			// link me in as previous sibling of old first child if there was one
					theMenu->previousSibling->nextSibling = theMenu;			// link me in as next sibling
				}
				else
				{
					theMenu->previousSibling = theMenu;						// loop previous sibling to myself
					theMenu->nextSibling = theMenu;							// loop next sibling to myself
				}
				baseMenu->firstChild = theMenu;								// this is now the first child
			}
			else							// creating a child of the universe
			{
				if (firstMenu)
				{
					theMenu->nextSibling = firstMenu;
					theMenu->previousSibling = firstMenu->previousSibling;
					theMenu->previousSibling->nextSibling = theMenu;
					firstMenu->previousSibling = theMenu;
				}
				else
				{
					theMenu->previousSibling = theMenu;
					theMenu->nextSibling = theMenu;
				}
				firstMenu = theMenu;
			}
			break;
		}
		case MR_LASTCHILD:
		{
			theMenu->parent = baseMenu;										// set parent
			if (baseMenu)													// see if linking as a child of a current menu
			{
				if ((theMenu->nextSibling = baseMenu->firstChild))
				{
					theMenu->previousSibling = theMenu->nextSibling->previousSibling;
					theMenu->nextSibling->previousSibling = theMenu;			// link me in as previous sibling of old first child if there was one
					theMenu->previousSibling->nextSibling = theMenu;			// link me in as next sibling
				}
				else
				{
					theMenu->previousSibling = theMenu;						// loop previous sibling to myself
					theMenu->nextSibling = theMenu;							// loop next sibling to myself
					baseMenu->firstChild = theMenu;
				}
			}
			else							// creating a child of the universe
			{
				if (firstMenu)
				{
					theMenu->nextSibling = firstMenu;
					theMenu->previousSibling = firstMenu->previousSibling;

					theMenu->previousSibling->nextSibling = theMenu;

					firstMenu->previousSibling = theMenu;
				}
				else
				{
					theMenu->previousSibling = theMenu;
					theMenu->nextSibling = theMenu;
					firstMenu = theMenu;
				}
			}
			break;
		}
		default:
		{
			break;
		}
	}
}

static void UnlinkEditorMenuElement(EDITOR_MENU* theMenu)
// unlink theMenu from the menu tree
{
	if (theMenu->parent)
	{
		if (theMenu->parent->firstChild == theMenu)
		{
			if (theMenu->nextSibling != theMenu)
			{
				theMenu->parent->firstChild = theMenu->nextSibling;
			}
			else
			{
				theMenu->parent->firstChild = NULL;
			}
		}
	}
	else
	{
		if (firstMenu == theMenu)
		{
			if (theMenu->nextSibling != theMenu)
			{
				firstMenu = theMenu->nextSibling;
			}
			else
			{
				firstMenu = NULL;
			}
		}
	}
	theMenu->nextSibling->previousSibling = theMenu->previousSibling;
	theMenu->previousSibling->nextSibling = theMenu->nextSibling;
}

static void ExtractAttributeParameter(const char* parameterInput, UINT32* parameterIndex, char* parameterBuffer, UINT32 parameterBufferSize)
// extract the next parameter from parameterInput[parameterIndex] into parameter buffer
// if the parameter would be too long, just truncate it
// leave paramterIndex updated to point to the end of the given parameter, no matter
// how long it is
{
	UINT32 outIndex;
	char theChar;
	bool done;

	outIndex = 0;
	done = false;
	while ((theChar = parameterInput[*parameterIndex]) && !done)			// make sure not end of string, or end of parameter
	{
		if (theChar != '\\')
		{
			(*parameterIndex)++;
			if (outIndex < parameterBufferSize)
			{
				parameterBuffer[outIndex++] = theChar;
			}
			else
			{
				done = true;
			}
		}
		else
		{
			if (parameterInput[*parameterIndex + 1] == '\\')				// see if quoted separator (not a separator at all then)
			{
				(*parameterIndex) += 2;
				if (outIndex < parameterBufferSize)
				{
					parameterBuffer[outIndex++] = '\\';				// place separator into buffer
				}
				else
				{
					done = false;
				}
			}
			else
			{
				done = true;											// ran into end of parameter list
			}
		}
	}
	if (outIndex < parameterBufferSize)
	{
		if (outIndex >= 1)
		{
			if (parameterBuffer[outIndex - 1] == ' ')
			{
				parameterBuffer[outIndex - 1] = '\0';
			}
			else
			{
				parameterBuffer[outIndex] = '\0';						// terminate buffer
			}
		}
		else
		{
			parameterBuffer[outIndex] = '\0';							// terminate buffer
		}
	}
	else
	{
		if (parameterBufferSize)
		{
			parameterBuffer[parameterBufferSize - 1] = '\0';			// terminate buffer by writing over last character
		}
	}
}

static bool GetModifierMaskAndValue(char* theString, UINT32* modifierMask, UINT32* modifierValue)
// convert a string of modifiers into a mask, and value
// if there is a problem, return FALSE
{
	int	i;
	int theLength;
	bool fail;

	*modifierMask = 0;
	*modifierValue = 0;
	theLength = (int)strlen(theString);
	fail = false;
	if (theLength >= 11)				// must be an entry for each flag
	{
		theLength = Min(11, theLength);
		for (i = 0; (!fail) && (i < theLength); i++)
		{
			switch (theString[i])
			{
				case 'x':
				case 'X':
				{
					break;
				}
				case '0':
				{
					*modifierMask |= modifierFlagTable[i];
					break;
				}
				case '1':
				{
					*modifierMask |= modifierFlagTable[i];
					*modifierValue |= modifierFlagTable[i];
					break;
				}
				default:
				{
					SetError(localErrorFamily, errorMembers[BADMODIFIERS], errorDescriptions[BADMODIFIERS]);
					fail = true;
					break;
				}
			}
		}
	}
	else
	{
		fail = true;
		SetError(localErrorFamily, errorMembers[BADMODIFIERS], errorDescriptions[BADMODIFIERS]);
	}
	return (!fail);
}

static bool ExtractMenuInfoFromAttributes(const char* theAttributes, EDITOR_MENU* theMenu, bool* keyData, bool* separator)
// Attempt to fill the parts of theMenu's record that are described in its string
// Currently supports:
// \Kkey(mod)		specifies the <key> to access the given menu and which modifiers(l=caps lock, s=shift, c=control, 0-7 are additional modifiers such as alt, option, command, etc...)
// \S				specifies that this menu item is a separator
//					Bind flags		 lsc01234567
// \W				specifies that this menu item is the MDI window menu item
// if there is a problem, return FALSE
{
	char parameterBuffer[256];							// max size for parameter string, anything larger gets truncated!
	UINT32 length;
	UINT32 attributeIndex;
	char theChar;
	bool fail;

	fail = false;
	attributeIndex = 0;
	theMenu->keyCode = 0;
	theMenu->keyModifiers = 0;
	theMenu->keyMask = 0;
	(*keyData) = false;
	(*separator) = false;
	while (!fail && (theChar = theAttributes[attributeIndex]))
	{
		if (theChar != '\\')								// eat separators
		{
			ExtractAttributeParameter(theAttributes, &attributeIndex, &(parameterBuffer[0]), 256);
			switch (parameterBuffer[0])					// check command part
			{
				case 'K':
				{
					length = (UINT32)strlen(&(parameterBuffer[1]));
					if (length >= 12)
					{
						if (!(fail = !GetModifierMaskAndValue(&(parameterBuffer[1]), &theMenu->keyMask, &theMenu->keyModifiers)))
						{
							if (!(fail = !EditorKeyNameToKeyCode(&(parameterBuffer[12]), &(theMenu->keyCode))))
							{
								(*keyData) = true;
							}
							else
							{
								SetError(localErrorFamily, errorMembers[BADKEYNAME], errorDescriptions[BADKEYNAME]);
							}
						}
					}
					else
					{
						fail = true;
						SetError(localErrorFamily, errorMembers[BADKEYNAME], errorDescriptions[BADKEYNAME]);
					}
					break;
				}
				case 'S':
				{
					(*separator) = true;
					break;
				}
				default:
				{
					break;
				}
			}
		}
		else
		{
			attributeIndex++;
		}
	}
	return (!fail);
}

void DeleteOSXMenuItem(EDITOR_MENU* theMenu)
//
//	Passed a menu item, delete the relevant OS X menu stuff
//
{
	if (theMenu->parent)							// top menu does not have a parent and should not be deleted
	{
		[theMenu->parent->menuHandle.submenu removeItem:theMenu->menuHandle];
	}
	else
	{
		[[NSApp mainMenu] removeItem:theMenu->menuHandle];
	}
}

bool CreateOSXMenuItem(EDITOR_MENU* baseMenu, EDITOR_MENU* theMenu, int menuRelationship, const char* theAttributes, bool active)
// Create a menu item
{
	bool result;

	result = false;
	if (ExtractMenuInfoFromAttributes(theAttributes, theMenu, &theMenu->hasKeyData, &theMenu->separator))
	{
		if (theMenu->separator)
		{
			theMenu->menuHandle = [NSMenuItem separatorItem];
		}
		else
		{
			theMenu->menuHandle = [[NSMenuItem alloc] initWithTitle:[NSString stringWithCString:theMenu->itemText encoding:NSASCIIStringEncoding] action:nil keyEquivalent:@""];
		}
		theMenu->menuID = globalMenuID++;														// each menu has a unique menu ID
		theMenu->menuHandle.tag = theMenu->menuID;
		NSEventModifierFlags menuModifiers = 0;
		if (theMenu->hasKeyData)
		{
			if ((theMenu->keyModifiers & EEM_MOD0))
			{
				menuModifiers |= NSCommandKeyMask;
			}
			if ((theMenu->keyModifiers & EEM_MOD1))
			{
				menuModifiers |= NSAlternateKeyMask;
			}
			if ((theMenu->keyModifiers & EEM_CTL))
			{
				menuModifiers |= NSControlKeyMask;
			}
			if ((theMenu->keyModifiers & EEM_SHIFT))
			{
				menuModifiers |= NSShiftKeyMask;
			}
			unichar c = g_virtKeyToKeyEquivalentArray[theMenu->keyCode];
			NSString* key = [[NSString stringWithCharacters:&c length:1] lowercaseString];

			[theMenu->menuHandle setKeyEquivalent:key];
			[theMenu->menuHandle setKeyEquivalentModifierMask:menuModifiers];
		}
		theMenu->menuHandle.enabled = active;
		[theMenu->menuHandle setTag:theMenu->menuID];
		[theMenu->menuHandle setTarget:menuHandler];
		[theMenu->menuHandle setAction:@selector(handleMenuEvent:)];

		if (baseMenu)
		{
			switch (menuRelationship)
			{
				case MR_NEXTSIBLING:
				case MR_PREVIOUSSIBLING:
				{
					if (baseMenu->menuHandle.submenu == nil)
					{
						baseMenu->menuHandle.submenu = [[NSMenu alloc] initWithTitle:[NSString stringWithCString:baseMenu->itemText encoding:NSASCIIStringEncoding]];
					}
					if (menuRelationship == MR_PREVIOUSSIBLING)
					{
						[baseMenu->menuHandle.submenu insertItem:theMenu->menuHandle atIndex:0];
					}
					else
					{
						[baseMenu->menuHandle.submenu addItem:theMenu->menuHandle];
					}
					result = true;
					break;
				}
				case MR_FIRSTCHILD:
				case MR_LASTCHILD:
				{
					if (baseMenu->menuHandle.submenu == nil)
					{
						baseMenu->menuHandle.submenu = [[NSMenu alloc] initWithTitle:[NSString stringWithCString:baseMenu->itemText encoding:NSASCIIStringEncoding]];
					}
					if (menuRelationship == MR_FIRSTCHILD)
					{
						[baseMenu->menuHandle.submenu insertItem:theMenu->menuHandle atIndex:0];
					}
					else
					{
						[baseMenu->menuHandle.submenu addItem:theMenu->menuHandle];
					}
					break;
				}
			}
			result = true;
		}
		else
		{
			if (!theMenu->separator)										// can't have a separator as top menu item
			{
				if (menuRelationship == MR_FIRSTCHILD)
				{
					[[NSApp mainMenu] insertItem:theMenu->menuHandle atIndex:0];
				}
				else
				{
					[[NSApp mainMenu] addItem:theMenu->menuHandle];
				}
				result = true;
			}
			else
			{
				result = true;
			}
		}
	}
	return (result);
}

static EDITOR_MENU* CreateEditorMenuAtBase(EDITOR_MENU* baseMenu, int menuRelationship, const char* theName, const char* theAttributes, const char* dataText, bool active)
// create a new menu with theName, as menuRelationship to baseMenu
// and return a pointer to it
// if baseMenu is NULL it indicates the top of the hierarchy
// therefore, menuRelationship must be MR_FIRSTCHILD, or MR_LASTCHILD as it is not possible to
// create menus which are siblings of the root
// theAttributes is a machine specific string that describes machine specific attributes of the menu
// if active is FALSE, the menu should be created inactive
// if there is a problem, set the error, and return NULL
{
	EDITOR_MENU* theMenu;

	if ((theMenu = (EDITOR_MENU*)MNewPtrClr(sizeof(EDITOR_MENU))))						// create editor menu element
	{
		if ((theMenu->dataText = (char*)MNewPtr((UINT32)(strlen(dataText) + 1))))
		{
			strcpy(theMenu->dataText, dataText);											// copy the data text for this menu
			if ((theMenu->itemText = (char*)MNewPtr((UINT32)(strlen(theName) + 1))))
			{
				strcpy(theMenu->itemText, theName);										// copy the name
				if (CreateOSXMenuItem(baseMenu, theMenu, menuRelationship, theAttributes, active))
				{
					LinkEditorMenuElement(theMenu, baseMenu, menuRelationship);
					return (theMenu);
				}
				MDisposePtr(theMenu->itemText);
			}
			MDisposePtr(theMenu->dataText);
		}
		MDisposePtr(theMenu);
	}
	return (NULL);
}

char* GetEditorMenuDataText(EDITOR_MENU* theMenu)
// return a pointer to the data text string for theMenu
{
	return (theMenu->dataText);
}

bool SetEditorMenuDataText(EDITOR_MENU* theMenu, const char* theDataText)
// set the data text for theMenu
// if this fails, leave the data text, set the error, and return FALSE
{
	char* newDataText;

	if ((newDataText = (char*)MNewPtr((UINT32)(strlen(theDataText) + 1))))
	{
		strcpy(newDataText, theDataText);
		MDisposePtr(theMenu->dataText);
		theMenu->dataText = newDataText;
		return (true);
	}
	return (false);
}

EDITOR_MENU* CreateEditorMenu(int argc, const char* argv[], int menuRelationship, const char* name, const char* attributes, const char* dataText, bool active)
// create a new menu with name, as menuRelationship to the passed argc,argv path
// the path is absolute within the menu hierarchy
// attributes is a machine specific string that describes machine specific attributes of the menu
// if active is FALSE, the menu should be created inactive
// if there is a problem, set the error, and return FALSE
{
	EDITOR_MENU* parentMenu;
	EDITOR_MENU* testMenu;
	bool fail;

	parentMenu = NULL;									// start at the top
	if (GetEditorMenu(argc, argv, &parentMenu))			// locate item to link this one to
	{
		testMenu = parentMenu;
		fail = false;
		if (menuRelationship == MR_NEXTSIBLING || menuRelationship == MR_PREVIOUSSIBLING)	// see if hunting for sibling
		{
			if (parentMenu)								// if he wants sibling, better not be at top
			{
				testMenu = testMenu->parent;				// back up one
				if (GetEditorMenu(1, &name, &testMenu))	// see if an item already exists with this name in the place we are putting it
				{
					if (testMenu != parentMenu)			// cannot delete the one we want to add to!!
					{
						DisposeEditorMenu(testMenu);	// kill it before adding new item
					}
					else
					{
						SetError(localErrorFamily, errorMembers[BADRELATIONSHIP], errorDescriptions[BADRELATIONSHIP]);
						fail = true;
					}
				}
			}
			else
			{
				SetError(localErrorFamily, errorMembers[BADRELATIONSHIP], errorDescriptions[BADRELATIONSHIP]);
				fail = true;
			}
		}
		else
		{
			if (GetEditorMenu(1, &name, &testMenu))			// see if an item already exists with this name in the place we are putting it
			{
				DisposeEditorMenu(testMenu);				// kill it before adding new item
			}
		}
		if (!fail)
		{
			return (CreateEditorMenuAtBase(parentMenu, menuRelationship, name, attributes, dataText, active));
		}
	}
	else
	{
		SetError(localErrorFamily, errorMembers[BADPATH], errorDescriptions[BADPATH]);
	}
	return (NULL);
}

void DisposeEditorMenu(EDITOR_MENU* theMenu)
// remove theMenu (and any children it may have)
// if NULL is passed, dispose of ALL menus
// NOTE: this calls itself recursively
//
{
	EDITOR_MENU* subMenu;

	while ((subMenu = GetRelatedEditorMenu(theMenu, MR_LASTCHILD)))		// see if this one has any children
	{
		DisposeEditorMenu(subMenu);
	}
	if (theMenu)
	{
		DeleteOSXMenuItem(theMenu);
		UnlinkEditorMenuElement(theMenu);
		MDisposePtr(theMenu->itemText);
		MDisposePtr(theMenu->dataText);
		MDisposePtr(theMenu);
	}
}

void UninitMenuManager()
{
	DisposeEditorMenu(NULL);										// get rid of any menus that were created
}

bool InitMenuManager()
// Initialize the menu manager.
{
	firstMenu = NULL;
	globalMenuID = 4096;												// each menu needs an ID, start here
	menuHandler = [[MenuObjCObject alloc] init];

	NSMenuItem* e93Menu = [[NSApp mainMenu] itemWithTag:1];
	NSMenuItem* menuItem;
	menuItem = [e93Menu.submenu itemWithTag:2];
	[menuItem setTarget:menuHandler];
	[menuItem setAction:@selector(handleMenuEvent:)];
	menuItem = [e93Menu.submenu itemWithTag:3];
	[menuItem setTarget:menuHandler];
	[menuItem setAction:@selector(handleMenuEvent:)];

	return (true);
}
