// Misc low level stuff
// Copyright (C) 2009 Fred Allen.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"
#import <AppKit/AppKit.h>

// routines callable by the editor

const char* GetEditorLocalVersion()
// return a pointer to a constant string, that tells something about this local
// version of the editor (used by the "version" command)
// it is ok to return NULL
{
	return ("OS X Cocoa GUI Version 1.0");
}

bool PointInRECT(INT32 x, INT32 y, EDITOR_RECT* rect)
// return true if x, and y fall within rect
{
	if ((x >= rect->x) && (y >= rect->y))
	{
		if ((x < (rect->x + (int)rect->w)) && (y < (rect->y + (int)rect->h)))
		{
			return (true);
		}
	}
	return (false);
}

void EditorBeep()
// make a noise so we can tell what is going on
{
	NSBeep();
}

void ReportMessage(const char* format, ...)
// report messages to the console
{
	va_list args;
	char tempString[256];

	va_start(args, format);
	vsnprintf(tempString, sizeof(tempString), format, args);	// make the string
	va_end(args);

	OkDialog(tempString);
}

void ShowBusy()
// the editor calls this when it is doing something that may take some
// time, it should show the user (by changing the pointer or something)
// this call is nestable
{
// nothing to do here, OS X takes care of this for us
}

void ShowNotBusy()
// when the editor is again ready to accept user input, it will call this
// it should undo anything ShowBusy did
// every ShowBusy will be exactly balanced by a call to ShowNotBusy
{
}

char* CreateWindowTitleFromPath(const char* absolutePath)
// use absolute path to generate a string (which is returned) that
// should be placed in the title of a window
// the string should be disposed of using MDisposePtr
// if there is a problem, SetError, and return NULL
{
	int pathLength;
	int nameLength;
	char* returnString;

	pathLength = (int)strlen(absolutePath);
	if ((returnString = (char*)MNewPtr(pathLength + 5)))
	{
		nameLength = 0;
		while (pathLength && absolutePath[pathLength - 1] != PATH_SEP)	// move backwards until the first part of the path is encountered
		{
			nameLength++;
			pathLength--;
		}
		sprintf(returnString, "%.*s -- %.*s", nameLength, &(absolutePath[pathLength]), pathLength, &(absolutePath[0]));
		return (returnString);
	}
	return (NULL);
}

char* realpath2(const char* path, char* outputPath)
// This routine replaces the system call "realpath" since there
// still exist machines that do not implement it
// This function directly replaces "realpath"
// NOTE: if path is passed as a string of length 0, this
// will return the current directory (which is what the real
// realpath does)
{

#define	MAX_LINK_DEPTH	256						// maximum number of links we will follow before giving up

	int currentLinkPath;						// tells which linkBuffer to read into next time we readlink
	char linkBuffer[2][MAXPATHLEN];				// two buffers to read links into
	char* linkPath;								// pointer to current link buffer
	int linkDepth;								// number of times links have been traversed
	int outputLength;							// current length of output buffer
	int linkLength;								// length of link returned by readlink
	bool hadRelativeComponent;					// boolean to tell if ., or .. was seen

	currentLinkPath = 0;							// use buffer 0 next time
	linkDepth = 0;								// so far, no links encountered
	outputLength = 0;								// current length of output path

	if (*path != PATH_SEP)							// if path does not start with PATH_SEP it is relative to current directory (even if path[0]=='\0')
	{
		if (getcwd(outputPath, MAXPATHLEN))		// get absolute path up to this point
		{
			outputLength = (int)strlen(outputPath);	// get the length of the path returned
			if ((!outputLength) || outputPath[outputLength - 1] != PATH_SEP)	// make sure the end of the path has a PATH_SEP on it
			{
				outputPath[outputLength++] = PATH_SEP;	// drop it in
			}
		}
		else
		{
			errno = ENOENT;						// something went amiss with getcwd, so make up an error code, and leave
			return (NULL);
		}
	}
	else
	{
		outputPath[outputLength++] = PATH_SEP;	// path is absolute, so start it
		path++;									// eat the PATH_SEP in the input
	}

	while (*path)								// get each component of the path, and expand
	{
		if (*path != PATH_SEP)						// gobble up separators
		{
			hadRelativeComponent = false;			// assume no ., or ..
			if (path[0] == '.')					// see if possible relative path component
			{
				if (path[1] == '.')
				{
					if (!path[2] || path[2] == PATH_SEP)	// just got '..', so move back over it, and previous path component, if any
					{
						path += 2;				// skip over '..' of input
						if (outputLength > 1)		// if at root, do not backup
						{
							outputLength--;		// step back over trailing PATH_SEP
							while (outputLength && outputPath[outputLength - 1] != PATH_SEP)
							{
								outputLength--;
							}
						}
						hadRelativeComponent = true;	// tell rest of loop this was handled
					}
				}
				else
				{
					if (!path[1] || path[1] == PATH_SEP)	// just saw '.', which is redundant, so remove it
					{
						path++;						// skip over '.'
						hadRelativeComponent = true;	// tell rest of loop this was handled
					}
				}
			}

			if (!hadRelativeComponent)				// if the component was not relative, copy it to the output, and check for links
			{
				while (*path && *path != PATH_SEP)		// copy the next pathname component
				{
					if (outputLength < MAXPATHLEN - 1)	// check for overflow, leave room for 0 termination
					{
						outputPath[outputLength++] = *path++;	// copy the character
					}
					else
					{
						errno = ENAMETOOLONG;			// ran out of space during the copy, complain
						return (NULL);
					}
				}

				outputPath[outputLength] = '\0';		// terminate the path that is being created
				linkPath = linkBuffer[currentLinkPath];	// point to the spare link buffer
				if ((linkLength = (int)readlink(outputPath, linkPath, MAXPATHLEN - 1)) >= 0)	// attempt to see what this is linked to, MAXPATHLEN-1 is to leave room for 0 terminator
				{
					if (linkDepth++ <= MAX_LINK_DEPTH)	// make sure we have not exhausted max number of links
					{
						linkPath[linkLength] = '\0';	// terminate the path returned by readlink
						if (linkPath[0] == PATH_SEP)	// see if the link is absolute or relative
						{
							outputLength = 0;			// if the link is absolute, then the output is completely re-written, and scan continues from this point
						}
						else
						{
							while (outputLength && outputPath[outputLength - 1] != PATH_SEP)	// the link is relative, so remove the link component from the output
							{
								outputLength--;
							}
							if (outputLength)		// step back over the path separator too
							{
								outputLength--;
							}
						}
						if (linkLength + strlen(path) < MAXPATHLEN)	// make sure the remainder of the path will fit into the buffer
						{
							strcat(linkPath, path);	// append what's left of the path to the output of readlink
							path = linkPath;			// call this the new path
							currentLinkPath ^= 1;		// swap link buffers
						}
						else
						{
							errno = ENAMETOOLONG;
							return (NULL);
						}
					}
					else
					{
						errno = ELOOP;				// too many links traversed, complain
						return (NULL);
					}
				}
				else
				{
					if (errno != EINVAL)				// anything but EINVAL is fatal, EINVAL is ok, the file is just not a symbolic link
					{
						return (NULL);
					}
				}
				outputPath[outputLength++] = PATH_SEP;	// add separator to the output
			}
		}
		else
		{
			path++;									// skip extraneous path separators
		}
	}

	if ((outputLength > 1) && outputPath[outputLength - 1] == PATH_SEP)	// realpath does not return the last PATH_SEP unless we're at the root
	{
		outputLength--;								// strip off final PATH_SEP
	}
	outputPath[outputLength] = '\0';					// terminate the result
	return (outputPath);								// return it
}

char* CreateAbsolutePath(const char* relativePath)
// given a relative path, attempt to make it an absolute path
// NOTE: if the relativePath does not point to anything valid, or the absolute path buffer
// could not be allocated, SetError, and return NULL
// the returned buffer will be 0 terminated, and must be freed by calling
// MDisposePtr
// NOTE: the result of calling this for the same file MUST always be
// EXACTLY the same string (case sensitive). The editor may perform
// strcmp style operations on these.
{
	char resolvedPath[MAXPATHLEN];
	char* returnPath;

	if (realpath2(relativePath, &(resolvedPath[0])))				// get real path if possible
	{
		if ((returnPath = (char*)MNewPtr((UINT32)(strlen(&(resolvedPath[0])) + 1))))	// create buffer to pass it back in
		{
			strcpy(returnPath, &(resolvedPath[0]));				// copy it into the buffer
			return (returnPath);
		}
	}
	else
	{
		SetStdCLibError();
	}
	return (NULL);
}

#define	STARTUP_NAME		"e93lib/e93rc.tcl"					// name of the startup file

static bool ScriptExists(const char* scriptPath)
// See if the script at script path exists, and is readable
// If so, return true, if not, return false
{
	int fileID;

	if ((fileID = open(scriptPath, O_RDONLY)) >= 0)
	{
		close(fileID);
		return (true);
	}
	return (false);
}

bool LocateStartupScript(char* scriptPath)
// Search for the editor's startup script, return a path
// to the script which is located.
// NOTE: scriptPath must point to at least MAXPATHLEN bytes
// If no script could be located, SetError and return false
{
	sprintf(scriptPath, "%s", STARTUP_NAME);
	if (ScriptExists(scriptPath))
	{
		return (true);
	}
	SetError("Failed to locate '%s'", scriptPath);
	return (false);
}

EDITOR_FILE* OpenEditorReadFile(const char* path)
// open a document file for reading into the editor
// if there is an error, SetError will be called, and NULL returned
{
	EDITOR_FILE* editorFile;

	if ((editorFile = (EDITOR_FILE*)MNewPtr(sizeof(EDITOR_FILE))))
	{
		if ((editorFile->fileHandle = open(path, O_RDONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) != -1)
		{
			return (editorFile);
		}
		else
		{
			SetStdCLibError();						// set high level error information
		}
		MDisposePtr(editorFile);
	}
	return (NULL);
}

EDITOR_FILE* OpenEditorWriteFile(const char* path)
// open a document file for writing into by the editor
// if there is an error, SetError will be called, and NULL returned
{
	EDITOR_FILE* editorFile;

	if ((editorFile = (EDITOR_FILE*)MNewPtr(sizeof(EDITOR_FILE))))
	{
		if ((editorFile->fileHandle = open(path, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) != -1)
		{
			return (editorFile);
		}
		else
		{
			SetStdCLibError();						// set high level error information
		}
		MDisposePtr(editorFile);
	}
	return (NULL);
}

bool ReadEditorFile(EDITOR_FILE* file, UINT8* buffer, UINT32 bytesToRead, UINT32* bytesRead)
// read at most bytesToRead from file into buffer
// return the actual amount read in bytesRead
// if bytesRead is less than bytesToRead, then EOF is assumed
// if there is an error, SetError will be called, and false returned
{
	int bytesJustRead;

	if ((bytesJustRead = (int)read(file->fileHandle, (char*)buffer, (int)bytesToRead)) != (int)bytesToRead)
	{
		if (bytesJustRead == -1)			// if there was an error, set it it
		{
			SetStdCLibError();			// set high level error information
			*bytesRead = 0;
			return (false);
		}
	}
	*bytesRead = bytesJustRead;
	return (true);
}

bool WriteEditorFile(EDITOR_FILE* file, UINT8* buffer, UINT32 bytesToWrite, UINT32* bytesWritten)
// write bytesToWrite from buffer into file
// return the actual amount written in bytesWritten
// bytesWritten should be the same as bytesToWrite, unless there was an error
// if there is an error, SetError will be called, and false returned
{
	int bytesJustWritten;

	if ((bytesJustWritten = (int)write(file->fileHandle, (char*)buffer, (int)bytesToWrite)) != (int)bytesToWrite)
	{
		if (bytesJustWritten == -1)			// if there was an error, set it it
		{
			SetStdCLibError();				// set high level error information
			*bytesWritten = 0;
			return (false);
		}
	}
	*bytesWritten = bytesJustWritten;
	return (true);
}

void CloseEditorFile(EDITOR_FILE* file)
// when the editor is finished with a file, it will call this
{
	close(file->fileHandle);					// get rid of the file
	MDisposePtr(file);
}

void SleepMilliSecond(UINT32 mSec)
//	Sleep for mSec milliseconds
{
	struct timespec theTimeSpec;

	theTimeSpec.tv_sec = mSec / 1000000;
	theTimeSpec.tv_nsec = (mSec % 1000000) * 1000000;
	nanosleep(&theTimeSpec, NULL);
}

UINT32 GetMilliSecondTicks()
//	Get millisecond ticks
{
	struct timeval ticks;

	gettimeofday(&ticks, NULL);
	return (UINT32)((ticks.tv_sec * 1000) + (ticks.tv_usec / 1000));
}

